CREATE TABLE species(name VARCHAR(20), clade VARCHAR(20), legs INT, extinct BOOL);
INSERT INTO species(name, clade, legs, extinct) VALUES ('human', 'mammal', 2, FALSE);
INSERT INTO species(name, clade, legs, extinct) VALUES ('diplodocus', 'reptile', 4, TRUE);
INSERT INTO species(name, clade, legs, extinct) VALUES ('bumblebee', 'insect', 6, FALSE);
INSERT INTO species(name, clade, legs, extinct) VALUES ('leech', 'annelid', 0, FALSE);
